@define-color theme_base_color #333333;
@define-color theme_bg_color #393f3f;

@define-color entry_tag_bg #888a85;
@define-color entry_tag_fg black;
@define-color photos_collection_bg #000000;

GdMainIconView.content-view {
    -GdMainIconView-icon-size: 40;
}

/* Make spinner visible on both dark and bright backgrounds w/o making
 * it look ugly/weird.
 */
GdMainIconView.content-view.cell:active {
    color: gray;
}

.documents-counter {
    background-image: url("dnd-counter.svg");
    background-size: contain;
    background-color: transparent;
    color: @theme_base_color;
    font-weight: bold;
}

.photos-entry-tag {
    background-image: none;
    background-color: @entry_tag_bg;
    color: @entry_tag_fg;

    border-radius: 4px;
    border-width: 0;

    margin: 2px;
    padding: 4px;
}

.photos-entry-tag:hover {
    background-color: shade(@entry_tag_bg, 1.10);
    color: @entry_tag_fg;
}

.photos-entry-tag.button,
.photos-entry-tag.button:hover,
.photos-entry-tag.button:active,
.photos-entry-tag.button:active:hover {
    background-color: transparent;
    background-image: none;
    border-image: none;
    border-width: 0;
}

.photos-entry-tag.button:hover {
    color: shade(@entry_tag_bg, 2.10);
}

.documents-scrolledwin.frame {
    border-width: 1px 0 0;
    border-radius: 0;
}

.photos-icon-bg {
  -gtk-icon-shadow: 0 1px #000000;
}

.photos-fade-in {
    opacity: 1.0;
    transition: opacity 0.2s ease-out;
}

.photos-fade-out {
    opacity: 0.0;
    transition: opacity 0.2s ease-out;
}

.photos-collection-icon {
    background-color: @photos_collection_bg;
    border-radius: 8px;
}
