/*
 * Photos - access, organize and share your photos on GNOME
 * Copyright © 2014 – 2017 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef PHOTOS_ICONS_H
#define PHOTOS_ICONS_H

#include <glib.h>

G_BEGIN_DECLS

#define PHOTOS_ICON_CONTENT_LOADING_SYMBOLIC "content-loading-symbolic"

#define PHOTOS_ICON_EDIT_FIND_SYMBOLIC "edit-find-symbolic"

#define PHOTOS_ICON_FAVORITE "starred"
#define PHOTOS_ICON_FAVORITE_SYMBOLIC PHOTOS_ICON_FAVORITE "-symbolic"

#define PHOTOS_ICON_GO_NEXT_SYMBOLIC "go-next-symbolic"
#define PHOTOS_ICON_GO_PREVIOUS_SYMBOLIC "go-previous-symbolic"

#define PHOTOS_ICON_IMAGE_AUTO_ADJUST_SYMBOLIC "image-auto-adjust-symbolic"
#define PHOTOS_ICON_IMAGE_ADJUST_COLOR_SYMBOLIC "image-adjust-color-symbolic"
#define PHOTOS_ICON_IMAGE_CROP_SYMBOLIC "image-crop-symbolic"
#define PHOTOS_ICON_IMAGE_EDIT_SYMBOLIC "image-edit-symbolic"
#define PHOTOS_ICON_IMAGE_FILTER_SYMBOLIC "image-filter-symbolic"
#define PHOTOS_ICON_IMAGE_SHARE_SYMBOLIC "emblem-shared-symbolic"

#define PHOTOS_ICON_IMAGE_X_GENERIC_SYMBOLIC "image-x-generic-symbolic"

#define PHOTOS_ICON_NOT_FAVORITE "non-starred"
#define PHOTOS_ICON_NOT_FAVORITE_SYMBOLIC PHOTOS_ICON_NOT_FAVORITE "-symbolic"

#define PHOTOS_ICON_OBJECT_SELECT_SYMBOLIC "object-select-symbolic"

#define PHOTOS_ICON_PHOTOS_SYMBOLIC "emblem-photos-symbolic"

#define PHOTOS_ICON_PROCESS_STOP_SYMBOLIC "process-stop-symbolic"

#define PHOTOS_ICON_SYSTEM_SYMBOLIC "open-menu-symbolic"

#define PHOTOS_ICON_VIDEO_DISPLAY_SYMBOLIC "video-display-symbolic"

#define PHOTOS_ICON_WINDOW_CLOSE_SYMBOLIC "window-close-symbolic"

G_END_DECLS

#endif /* PHOTOS_ICONS_H */
